"""A Node is an object which autmatically manages it's owner(s).

A Node should alweys be stored in attribute members declared as
NodeReference or as member of a NodeMap or NodeList."""


import weakref

class Node(object):

    def __init__(self):
        self._owners = []

    def _addOwner(self,owner):
        self._owners.append(weakref.ref(owner,self._delRef))

    def _delOwner(self,owner):
        for i in range(len(self._owners)):
            if self._owners[i]() is owner:
                del self._owners[i]
                return

    def _delRef(self,ref):
        for i in range(len(self._owners)):
            if self._owners[i] is ref:
                del self._owners[i]
                return

    def _owners(self):
        return ( x() for x in self._owners )

    def _owner(self):
        if self._owners and self._owners[0]() is not None:
            return self._owners[0]()
        return None

    owners = property(_owners, None, None, None)
    owner = property(_owner, None, None, None)

id_index = 0
def _make_id():
    global id_index
    id_index += 1
    return '__node_elt_%d' % id_index

def NodeReference(id=None):
    if id is None : id = _make_id()
    return property(lambda self, id=id: getattr(self,id,None),
                    lambda self, node, id=id: _NodeReference_set(self,node,id),
                    None,
                    None)

def _NodeReference_set(self, node, id):
    old = getattr(self,id,None)
    if old is not None : old._delOwner(self)
    setattr(self,id,node)
    if node is not None : node._addOwner(self)


class _NodeMap(dict):

    def __init__(self, owner):
        dict.__init__(self)
        self._owner = weakref.ref(owner)

    def __setitem__(self, key, node):
        try: v = self[key]
        except KeyError: pass
        else: v._delOwner(self._owner())
        dict.__setitem__(self, key, node)
        node._addOwner(self._owner())

    def __delitem__(self, key):
        try: v = self[key]
        except KeyError: pass
        else: v._delOwner(self._owner())
        dict.__delitem__(self,key)

    def clear(self):
        owner = self._owner()
        for v in self.itervalues() : v._delOwner(owner)
        dict.clear(self)

    def setdefault(self, key, node):
        if not self.has_key(key) : self[key] = node

    def pop(self, key):
        v = self[key]
        del self[key]
        return v

    def popitem(self):
        try: k = self.iterkeys().next()
        except StopIteration: raise KeyError, 'popitem(): dictionary is empty'
        return (k, self.pop(k))

    def update(self, *args, **kws):
        for k,v in dict(*args,**kws).iteritems() : self[k] = v

def NodeMap(id=None):
    if id is None : id = _make_id()
    return property(lambda self, id=id : _NodeMap_get(self, id),
                    lambda self, elts, id=id : _NodeMap_set(self, elts, id),
                    None)

def _NodeMap_get(self, id):
    if not hasattr(self, id) : setattr(self, id, _NodeMap(self))
    return getattr(self, id)

def _NodeMap_set(self, elts, id):
    if hasattr(self, id) : getattr(self, id).clear()
    else                 : setattr(self, id, _NodeMap(self))
    getattr(self, id).update(elts)
    

class _NodeList(list):

    def __init__(self, owner):
        list.__init__(self)
        self._owner = weakref.ref(owner)

    def __setitem__(self, index, node):
        if type(index) is not slice:
            self[index] # throw IndexError if index invalid
            index = slice(index,index+1,None)
            node = [ node ]
        owner = self._owner()
        for i in range(*index.indices(len(self))) : self[i]._delOwner(owner)
        try:
            list.__setitem__(self, index, node)
        except ValueError:
            for i in range(*index.indices(len(self))) : self[i]._addOwner(owner)
            raise
        else:
            for n in node : n._addOwner(owner)

    def __delitem__(self, index):
        if type(index) is not slice:
            self[index] # throw IndexError if index invalid
            index = slice(index,index+1,None)
        self.__setitem__(index,[])

    def __setslice__(self, i, j, seq):
        self[max(0, i):max(0, j):] = seq

    def __delslice__(self, i, j):
        del self[max(0, i):max(0, j):]

    def append(self, node):
        self[len(self):len(self)] = [node]

    def extend(self, nodes):
        self[len(self):len(self)] = nodes

    def insert(self, index, node):
        self[index:index] = [node]

    def pop(self, index=None):
        if index is None : index = len(self)-1
        x = self[index]
        del self[index]
        return x

    def remove(self, node):
        del self[self.index(node)]

def NodeList(id=None):
    if id is None : id = _make_id()
    return property(lambda self, id=id : _NodeList_get(self, id),
                    lambda self, elts, id=id : _NodeList_set(self, elts, id),
                    None)

def _NodeList_get(self, id):
    if not hasattr(self, id) : setattr(self, id, _NodeList(self))
    return getattr(self, id)

def _NodeList_set(self, elts, id):
    if hasattr(self, id) : del getattr(self, id)[:]
    else                 : setattr(self, id, _NodeList(self))
    getattr(self, id).extend(elts)
